#  Copyright (c) 1997-2012
#  Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Darmstadt, Germany)
#  http://www.polymake.org
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2, or (at your option) any
#  later version: http://www.gnu.org/licenses/gpl.txt.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#-------------------------------------------------------------------------------
#  $Project: polymake $$Id: Sketch.pm 10609 2012-02-27 16:11:43Z herr $

# Interface to Sketch.
#
# This file only provides the basic functionality.  Visualization of polymake's various object types
# triggers code implemented in apps/*/rules/sketch.rules.

package Sketch::File;

use Polymake::Struct (
   [ new => ';$' ],
   [ '$title' => '#1' ],
   '@geometries',
   [ '$unnamed' => '0' ],
);

sub append {
   my $self = shift;
   push @{$self->geometries}, @_;
   foreach (@_) {
      if (length($_->name)) {
	 defined($self->title ||= $_->name);
      } else {
	 $_->name="unnamed__" . ++$self->unnamed;
      }
   }
}

sub header {
   my ($self) = @_;
   my $who=$ENV{USER};
   my $when=localtime();
   my $title=defined($self->title || "unnamed");
   return <<".";
% polymake for $who
% $when
% $title

.
}

sub trailer {
   return <<".";
global { language tikz }

% EOF
.
}

sub new_file {
   my $self=shift;
   new Sketch::File(@_);
}

sub toString {
   my $self=shift;
   $self->header . join("", map { $_->toString } @{$self->geometries}) . $self->trailer;
}

##############################################################################################
#
#  Basis class for all graphical objects handled by sketch
#
package Sketch::PointSet;
use Polymake::Struct (
   [ new => '$' ],
   [ '$source' => '#1' ],
   [ '$name' => '#1 ->Name' ],
);

# name without whitespace and such
# notice that "-", "{", and "}" occur in names of simplices of a TRIANGULATION
sub id {
   my ($self)=@_;
   my $id=$self->name;
   $id =~ s/[\s\{\}-]+//g;
   return $id;
}

sub header {
   my ($self)=@_;
   my $id=$self->id;
   return <<".";
def poly_$id {
.
}

sub trailer {
   my ($self)=@_;
   my $id=$self->id;
   my $view_point= $self->source->ViewPoint;
   $view_point = join ",", @$view_point;
   my $view_direction= $self->source->ViewDirection;
   $view_direction = join ",", @$view_direction;
   my $view_up= $self->source->ViewUp;
   $view_up = join ",", @$view_up;
   return <<".";
}

put { view ( ($view_point), ($view_direction), [$view_up]  ) } {poly_$id}

.
}

sub pointsToString {
   my ($self)=@_;
   my $id = $self->id; # this is used to make all Sketch ids globally unique

   my $d= is_object($self->source->Vertices) ? $self->source->Vertices->cols : 3;
   my $style=$self->source->VertexStyle;
   my $point_flag= is_code($style) || $style !~ $Visual::hidden_re ? "show" : "hide";
   my $color=$self->source->VertexColor;
   my $color_flag= is_code($color) ? "show" : "hide";
   my $k=0;
   my $labels=$self->source->VertexLabels;
#---
   my $thickness=$self->source->VertexThickness;
   my $thickness_flag= is_code($thickness) ? "show" : "hide";
   my @vertexcolor = split(/ /, $self->source->VertexColor->toFloat);
   my $vcstring = join ",", @vertexcolor;

   my $text= <<".";

  special|\\definecolor{pointcolor_$id}{rgb}{ $vcstring }\n|[lay=under]
  special|\\tikzstyle{pointstyle_$id} = [fill=pointcolor_$id]\n|[lay=under]

.

   foreach (@{$self->source->Vertices}) {
# FIXME: the following is unnecessarily complicated (copied from JavaView)
      # animated contents have point indices instead of real coords
      my $point=ref($_) ? Visual::print_coords($_) : "$_".(" 0"x($d-1));
      $point =~ s/\s+/, /g;
#      if (defined($labels)) {
#	 $attr=' name="' . $labels->($k) . '"';
#      }
      $text .= <<".";
  def v$k ($point)
.
      ++$k;
   }
   $text .= "\n";
   for (my $i=0; $i<$k; ++$i) {
      $text .= <<".";
  dots [dotsize=$thickness pt, style=pointstyle_$id] (v$i)
.
   }      
#   my $th= $thickness_flag eq "hide" ? defined($thickness || 1)*$JavaView::default::points_thickness : $JavaView::default::points_thickness;
#   $text .= <<".";
#          <thickness>$th</thickness>
#.
#
#   if ($color_flag eq "hide" && defined($color)) {
#      my $c=$color->toInt;
#      $text .= <<".";
#          <color type="rgb">$c</color>
#.
#   }
#   my $show_labels= defined($labels) ? "show" : "hide";
#   $text .= <<".";
#          <labelAtt visible="$show_labels"/>
#        </points>
#.
#   if ($color_flag eq "show") {
#      $text .= <<".";
#        <colors>
#.
#      for (my $i=0; $i<$k; ++$i) {
#	 my $c=$color->($i)->toInt;
#	 $text .= <<".";
#          <c>$c</c>
#.
#      }
#      $text .= <<'.';
#        </colors>
#.
#   }
#   if ($thickness_flag eq "show") {
#      $text .= <<'.';
#        <thicknesses>
#.
#      for (my $i=0; $i<$k; ++$i) {
#         my $th=defined($thickness->($i) || 1)+0;
#	 $text .= <<".";
#          <th>$th</th>
#.
#      }
#      $text .= <<'.';
#        </thicknesses>
#.
#   }
   return $text;
}

sub toString {
   my ($self)=@_;
   $self->header . $self->pointsToString . $self->trailer;
}

##############################################################################################
#
#  Wire model (e.g. a graph)

package Sketch::Wire;
use Polymake::Struct (
   [ '@ISA' => 'PointSet' ],
);

sub linesToString {
   my ($self)=@_;
   my $id=$self->id;

   my $arrows=$self->source->ArrowStyle;
   my $style=$self->source->EdgeStyle;
   my $line_flag= is_code($style) || $style !~ $Visual::hidden_re ? "show" : "hide";
   my $thickness=$self->source->EdgeThickness;
   my $thickness_flag= is_code($thickness) ? "show" : "hide";
   my $color=$self->source->EdgeColor;
   my $color_flag= is_code($color) ? "show" : "hide";
   my $color = is_code($self->source->EdgeColor) ? $self->source->EdgeColor->($self->source->all_edges)->toFloat : $self->source->EdgeColor->toFloat;
   my @linecolor = split(/ /, $color);
   my $lcstring = join ",", @linecolor;
   my $lsstring = "color=linecolor_$id, thick";
   if($arrows==1) {
       $lsstring .= ", arrows = -stealth, shorten >= 1pt";
   } elsif($arrows==-1) {
       $lsstring .= ", arrows = stealth-, shorten >= 1pt";
   }
   my $text = <<".";

  special|\\definecolor{linecolor_$id}{rgb}{ $lcstring }\n|[lay=under]
  special|\\tikzstyle{linestyle_$id} = [$lsstring]\n|[lay=under]

.

   my $labels=$self->source->EdgeLabels;
   for (my $e=$self->source->all_edges; $e; ++$e) {
      next if is_code($style) && $style->($e) =~ $Visual::hidden_re;
      my $attr="";
      if (defined($labels)) {
	 $attr=' name="' . $labels->($e) . '"';
      }
      my $a=$e->[0]; my $b=$e->[1];
# FIXME: if the names of the points are adjusted above, change this, too
      $text .= <<".";
  line [line style=linestyle_$id] (v$a) (v$b)
.
   }
#
#   my $th=$thickness_flag eq "hide" ? defined($thickness || 1)*$JavaView::default::lines_thickness : $JavaView::default::lines_thickness;
#   $text .= <<".";
#          <thickness>$th</thickness>
#.
#
#   if ($color_flag eq "hide" && defined($color)) {
#      my $c=$color->toInt;
#      $text .= <<".";
#          <color type="rgb">$c</color>
#.
#   }
#   my $show_labels= defined($labels) ? "show" : "hide";
#   $text .= <<".";
#          <labelAtt visible="$show_labels"/>
#        </lines>
#.
#   if ($color_flag eq "show") {
#      $text .= <<'.';
#        <colors>
#.
#      for (my $e=$self->source->all_edges; $e; ++$e) {
#	 next if is_code($style) && $style->($e) =~ $Visual::hidden_re;
#	 my $c=$color->($e)->toInt;
#	 $text .= <<".";
#          <c>$c</c>
#.
#      }
#      $text .= <<'.';
#        </colors>
#.
#   }
#   if ($thickness_flag eq "show") {
#      $text .= <<'.';
#        <thicknesses>
#.
#      for (my $e=$self->source->all_edges; $e; ++$e) {
#	 next if is_code($style) && $style->($e) =~ $Visual::hidden_re;
#         my $th=defined($thickness->($e) || 1)+0;
#	 $text .= <<".";
#          <th>$th</th>
#.
#      }
#      $text .= <<'.';
#        </thicknesses>
#.
#   }
#   $text .= <<'.';
#      </lineSet>
#.
   return $text;
}

sub toString {
   my ($self)=@_;
   $self->header . $self->pointsToString . $self->linesToString . $self->trailer;
}


##############################################################################################
#
#  Solid 2-d or 3-d body
#
package Sketch::Solid;
use Polymake::Struct (
   [ '@ISA' => 'PointSet' ],
);

sub facesToString {
   my ($self)=@_;
   my $id = $self->id; # this is used to make all Sketch ids globally unique

   my $style=$self->source->FacetStyle;
   my $transp=$self->source->FacetTransparency;
   my $face_flag= $style !~ $Visual::hidden_re ? "show" : "hide";
   my $color=$self->source->FacetColor;
   my $color_flag= is_code($color) ? "show" : "hide";
   my $edge_style=$self->source->EdgeStyle;
   my $edge_flag= is_code($edge_style) || $edge_style !~  $Visual::hidden_re ? "show" : "hide";
   my $backface_flag= $face_flag eq "show" && $self->source->Closed && !defined($transp) ? "hide" : "show";
   my $k=0;
   my $labels=$self->source->FacetLabels;

   my @facetcolor = split(/ /, $self->source->FacetColor->toFloat);
   my $fcstring = join ",", @facetcolor;

   my @edgecolor = split(/ /, $self->source->EdgeColor->toFloat);
   my $egstring = join ",", @edgecolor;

   my $thickness=defined($self->source->EdgeThickness) ? $self->source->EdgeThickness : 1;

   my $transp=defined($self->source->FacetTransparency) ? $self->source->FacetTransparency : 0.8;

   my $text = <<".";

  special|\\definecolor{facetcolor_$id}{rgb}{ $fcstring }\n|[lay=under]
  special|\\definecolor{edgecolor_$id}{rgb}{ $egstring }\n|[lay=under]
  special|\\tikzstyle{facetstyle_$id} = [fill=facetcolor_$id, opacity=$transp]\n|[lay=under]
  special|\\tikzstyle{edgestyle_$id} = [draw=edgecolor_$id, line width=$thickness pt, line cap=round]\n|[lay=under]

.

   foreach my $facet(@{$self->source->Facets}) {
       $text .= "  line [line style=edgestyle_$id]";
       foreach my $vertex(@$facet) {
	   $text.=" (v$vertex)";
       }
       my $first_v=$facet->[0];
       $text.=" (v$first_v)";
       $text.="\n";
   }
   
  if ($face_flag eq "show"){
   foreach my $facet(@{$self->source->Facets}) {
       $text .= "  polygon [cull=false, fill style=facetstyle_$id]";
       foreach my $vertex(@$facet) {
	   $text.=" (v$vertex)";
       }
       $text.="\n";
   }
  }

   return $text;
}

sub toString {
   my ($self)=@_;
   return $self->header . $self->pointsToString . $self->facesToString . $self->trailer;
}

1

# Local Variables:
# c-basic-offset:3
# End:
